// animation.cpp
// =============

#include "stdafx.h"
#include "resource.h" 
#include "mmsystem.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include "animation.h"

// ------------------------------------------------------------------------------
// CDib
// ------------------------------------------------------------------------------

CDib::CDib ()
{
	pBmpInfoHdr  = NULL ;
	pPixels      = NULL ;
	pColourTable = NULL ;
	numOfColours = 0 ;
	width        = 0 ;
	height       = 0 ;
	bpp          = 0 ;
	hBmp         = NULL ;
	isDibSection = false ;
	isPalettized = false ;
	isResource   = false ;
	bgColour = 0 ;
	bgPixel = 0 ;
	bgWord = 0 ;

	bgTriple.rgbtRed = 0 ;
	bgTriple.rgbtBlue = 0 ;
	bgTriple.rgbtGreen = 0 ;
}

CDib::~CDib ()
{
	if ( pBmpInfoHdr != NULL )
	{
		if ( isDibSection && hBmp != NULL ) 
			::DeleteObject(hBmp) ;
	
		if ( ! isResource )
			free (pBmpInfoHdr) ;
	}
}

int CDib::getNumOfColours ()
{
	int numColours ;

	if ( pBmpInfoHdr == NULL )
		return 0 ;

	numColours = pBmpInfoHdr->biClrUsed ;
	
	if ( numColours == 0 )
		numColours = 256 ;

	return numColours ;
}

int CDib::getBpp ()
{
	int bits ;

	if ( pBmpInfoHdr == NULL )
		return 0 ;

	bits = pBmpInfoHdr->biBitCount ;

	if ( bits == 0 )
		bits = 8 ;

	return bits ;
}

RGBQUAD * CDib::getColourTablePtr ()
{
	if ( this->pBmpInfoHdr == NULL )
		return NULL ;

	int bits ;

	bits = getBpp() ;

	if ( bits == 8 )
		return (RGBQUAD*)((BYTE*)pBmpInfoHdr + sizeof(BITMAPINFOHEADER)) ;
	else
		return NULL ;

}

BYTE * CDib::getPixelPtr ()
{
	BYTE * pByte ;

	int    colours ;

	if (this->pBmpInfoHdr == NULL ) 
		return NULL ;

	if ( isDibSection )
		return pPixels ;
	else
	{
		colours = getNumOfColours() ;

		pByte = (BYTE*)pBmpInfoHdr + sizeof(BITMAPINFOHEADER) + colours*sizeof(RGBQUAD);

		return pByte ;
	}
}

int CDib::loadFromResource ( int resId , bool updateable )
{
	// loads a bitmap resource

	HRSRC hResource ;

	HGLOBAL hGlobal ;

	BITMAPINFOHEADER  *   lpBmpInfoHdr ;

	hResource = FindResource ( NULL , (LPCTSTR)resId , RT_BITMAP ) ;
	if ( hResource == NULL )
		return ERROR_COULD_NOT_FIND_RESOURCE ;
		

	hGlobal = LoadResource ( NULL , hResource ) ;
	if ( hGlobal == NULL )
		return ERROR_COULD_NOT_LOAD_RESOURCE ;

	lpBmpInfoHdr = (BITMAPINFOHEADER*)LockResource(hGlobal) ; // get a pointer to the
	                                                          // resource

	if ( lpBmpInfoHdr == NULL ) 
		return ERROR_COULD_NOT_LOCK_RESOURCE ;

	// If the bitmap has to be updateable, it needs to be copied into a
	// new memory area, because resources are write-protected.

	if ( updateable ) 
	{
		// work out size required 

	    int bytesRequired;
	    int storageWidth;
		int width, height , bpp ;
		int pixelSize ;

	    // work out number of bytes required

		width  = lpBmpInfoHdr->biWidth ;
		height = lpBmpInfoHdr->biHeight ;
		bpp    = lpBmpInfoHdr->biBitCount ;

	    bytesRequired = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) ;
	
		if ( bpp == 8 )
		bytesRequired += 256*sizeof(RGBQUAD);

	    // but wait ... there's more!
	
		if ( bpp == 16 )
			pixelSize = 2 ;
		else if ( bpp == 24 )
			pixelSize = 3 ;
		else
			pixelSize = 1 ;

		storageWidth = (width*pixelSize + 3) & ~3 ; // each line gets rounded up to 32-bit word boundary

	    bytesRequired += storageWidth * height ; 

		// try to allocate the space for it

		char * pSpace ;

		pSpace = (char *)malloc ( bytesRequired ) ;

		if ( pSpace == NULL )
			return ERROR_COULD_NOT_ALLOCATE_MEMORY ; // couldn't allocate the space
		
		memcpy ( pSpace , (char*)lpBmpInfoHdr , bytesRequired ) ;

		lpBmpInfoHdr = (LPBITMAPINFOHEADER)pSpace ;
	}

	if ( this->pBmpInfoHdr != NULL && ! this->isResource) 
		free (this->pBmpInfoHdr) ;
	
	this->pBmpInfoHdr  = lpBmpInfoHdr ;
	this->pPixels      = this->getPixelPtr() ;
	this->pColourTable = this->getColourTablePtr();
	this->numOfColours = this->getNumOfColours() ;
	this->width        = (short)lpBmpInfoHdr->biWidth ;
	this->height       = (short)lpBmpInfoHdr->biHeight ;
	this->bpp          = lpBmpInfoHdr->biBitCount ;
	if (this->bpp == 0 )
		this->bpp = 8 ;

	this->isDibSection = false ;
	this->isPalettized = false ;
	
	this->isResource = ! updateable ;

	return 0 ;

}

int CDib::setBgColour ( unsigned int rgb , CPalette * pPal )
{
	// rgb is a RGB colour value, i.e. 3 bytes, 8 bits each :
	//
	//     BLUE then GREEN then RED
	//
	// This needs to be converted to a pixel value, appropriate for
	// the dib.
	//
	this->bgColour = rgb ;

	if ( this->bpp == 8 )
	{
		// convert the colour to a colour-table index

		if ( pPal != NULL )
		{
			this->bgPixel = pPal->GetNearestPaletteIndex ( rgb ) ;
		}
		else if ( this->isPalettized ) 
			return ERROR_PALETTE_IS_REQUIRED ;  // can't work out colour index
		                                        // if there is no palette and
		                                        // image is palettized
		else
		{
			CPalette tempPalette ;
			
			this->makePalette ( &tempPalette ) ;

			this->bgPixel = tempPalette.GetNearestPaletteIndex ( rgb ) ;

		}
	}
	else if ( this->bpp == 16 )
	{
		this->bgWord = ((GetRValue(rgb)>>3)<<10) + 
			           ((GetGValue(rgb)>>3)<<5) + 
						(GetBValue(rgb)>>3);
	}
	else if ( this->bpp == 24 )
	{
		this->bgTriple.rgbtBlue = GetRValue(rgb) ;
		this->bgTriple.rgbtGreen = GetGValue(rgb) ;
		this->bgTriple.rgbtRed = GetRValue(rgb) ;
	}
	else
		return ERROR_UNKNOWN_COLOUR_DEPTH ;

	return 0 ;
}

int CDib::bgFill ( CRect * pRect )
{
	// fills the dib's pixel area with the background colour

	int startRow , endRow , startCol , endCol ;
	int storageWidth ;

	if ( pRect == NULL )
	{
		startRow = 0 ;
		endRow = this->height ;
		startCol = 0 ;
		endCol = this->width ;
	}
	else
	{
		startRow = pRect->top ;
		endRow = pRect->bottom ;
		startCol = pRect->left ;
		endCol = pRect->right ;
		if (startRow < 0 )
			startRow = 0 ;
		else if (startRow >= this->height )
			startRow = this->height - 1 ;
		if ( endRow <= 0 ) 
			endRow = 1 ;
		else if ( endRow > this->height ) 
			endRow = this->height ;
		if ( startCol < 0 ) 
			startCol = 0 ;
		else if ( startCol >= this->width )
			startCol = this->width - 1 ;
		if ( endCol <= 0 )
			endCol = 1 ;
		else if ( endCol > this->width )
			endCol = this->width ;

		if ( startRow >= endRow || startCol >= endCol )
		{
			return ERROR_RECTANGLE_OUT_OF_RANGE ;
		}
	}

	storageWidth = (width*bpp/8 + 3) & ~3 ; // each line gets rounded up to 32-bit word boundary

	if ( bpp == 8 )
	{
		BYTE * pPix ;
		
		pPix = pPixels + (this->height - endRow) * storageWidth ;

		int row ;

		for (row = startRow ; row < endRow ; row++)
		{
			memset (pPix + startCol , bgPixel , endCol - startCol  ) ;
			pPix += storageWidth ;
		} 
  
	}
	else if ( bpp == 16 )
	{
		// Two bytes per pixel

		WORD* pPix ;
		
		pPix = (WORD*)(pPixels + (this->height - endRow)*storageWidth);

		int row ;

		for (row = startRow ; row < endRow ; row++)
		{
			int i ;
			for ( i = startCol ; i < endCol ; i++ )
				pPix[i] = bgWord ;
			
			pPix = (WORD*)((char*)pPix + storageWidth) ;
		} 
  
	}
	else if ( bpp == 24 )
	{
		RGBTRIPLE* pPix ;
		
		pPix = (RGBTRIPLE*)(pPixels + (this->height - endRow)*storageWidth);

		int row ;

		for (row = startRow ; row < endRow ; row++)
		{
			int i ;
			for ( i = startCol ; i < endCol ; i++ )
				pPix[i] = bgTriple ;
			
			pPix = (RGBTRIPLE*)((char*)pPix + storageWidth) ;
		} 
	}
	else
		return ERROR_UNKNOWN_COLOUR_DEPTH ; // unsupported colour depth
  
    return 0 ;
}


void CDib::trimColours () 
{
	BITMAPINFO * pBmpInfo ;

	BYTE *       pBmpPixels ;

	LPRGBQUAD    pColourTable ;

	int          numColours ;

	if ( pBmpInfoHdr->biBitCount != 0 && pBmpInfoHdr->biBitCount != 8 )
		return ;

    pBmpInfo = (BITMAPINFO *)((char*)pBmpInfoHdr);

	numColours = pBmpInfoHdr->biClrUsed; // number of colours used

	if (numColours == 0)
	{
		numColours = 256 ; // if biClrUsed is zero, then the
						   // maximum number of colours are defined
						   // in the RGBQUAD table (i.e.256)
	}


    pBmpPixels = (BYTE *)pBmpInfoHdr + sizeof(BITMAPINFOHEADER) 
		                                 + numColours*sizeof(RGBQUAD) ;

	pColourTable = (LPRGBQUAD)((char*)pBmpInfoHdr + sizeof(BITMAPINFOHEADER)) ;

    int used[256] ;

    int i, j ;

    int pixVal ;

    // clear the "used" array to zero

    for (i = 0 ; i < 256 ; i++)
        used[i] = 0 ;

    int width, height , storageWidth ;

    width = pBmpInfoHdr->biWidth ;

    height = pBmpInfoHdr->biHeight ;

    storageWidth = (width + 3) & ~3 ;

    for ( j = 0 ; j < height ; j++)
    {
        for ( i = 0 ;  i < width ; i++)
        {
            pixVal = *pBmpPixels++ ;
            if (pixVal < 256)
                used[pixVal] = 1 ;
        }

        pBmpPixels += storageWidth - width ;
    }
          
    j = 0 ;

    for (i = 0 ; i < 256 ; i++)
    {
        if ( used[i] == 0 )
        {
            // this colour entry not used, so set it to zero

            pColourTable[i].rgbRed = 0 ;
            pColourTable[i].rgbGreen = 0 ;
            pColourTable[i].rgbBlue = 0 ;
            pColourTable[i].rgbReserved = 0 ;

        }
        else
            // this colour entry is used, so increment count of colours used
            j++ ;
    }
}

int CDib::makePicture  (int width, int height , int * colourArray )
{
	// Create a new 256-colour bitmap, filled in black. Colour table contains 
	// default window colours.

	if ( pBmpInfoHdr != NULL )
	{
		free (pBmpInfoHdr) ;
		pBmpInfoHdr = NULL ;
	}

    int storageWidth ;

	int bytesRequired ;

    // work out number of bytes required

	storageWidth = (width + 3) & ~3 ; // each line gets rounded up to 32-bit word boundary

	bytesRequired = sizeof(BITMAPINFOHEADER) + 256*sizeof(RGBQUAD) + storageWidth*height ;

    LPBITMAPINFOHEADER  lpBmpInfoHdr = (LPBITMAPINFOHEADER)(malloc(bytesRequired)) ;

	if ( lpBmpInfoHdr == NULL )
		return ERROR_COULD_NOT_ALLOCATE_MEMORY ;

    BITMAPINFO * lpBmpInfo = (BITMAPINFO *)((char*)lpBmpInfoHdr);

    LPRGBQUAD lpColourTable = (LPRGBQUAD)((char*)lpBmpInfoHdr + sizeof(BITMAPINFOHEADER)) ;
    
    BYTE *  lpBmpPixels = (BYTE *)lpBmpInfoHdr + sizeof(BITMAPINFOHEADER) 
		                      + 256*sizeof(RGBQUAD) ; 


    
    lpBmpInfoHdr->biSize = sizeof(BITMAPINFOHEADER);
    lpBmpInfoHdr->biWidth = width ;
    lpBmpInfoHdr->biHeight = height ;
    lpBmpInfoHdr->biPlanes = 1 ;
    lpBmpInfoHdr->biBitCount = 8 ;
    lpBmpInfoHdr->biCompression = BI_RGB;
    lpBmpInfoHdr->biSizeImage = 0 ;
    lpBmpInfoHdr->biXPelsPerMeter = 0 ;
    lpBmpInfoHdr->biYPelsPerMeter = 0 ;
    lpBmpInfoHdr->biClrUsed = 0 ;
	                              // for 256-colour images, implies colour table has
	                              // 256 entries.
	                              //
    lpBmpInfoHdr->biClrImportant = 0 ;

	int i , staticColours , paletteSize ;

	// set up colour table. First need to get static colours from system palette

	PALETTEENTRY sysPal[256] ;
	HWND hwnd = GetActiveWindow() ;            // need a device context for getting
		                                       // system palette
	HDC hdc = GetDC(hwnd);

	staticColours = GetDeviceCaps(hdc, NUMCOLORS) ; // should be 20
	paletteSize = GetDeviceCaps(hdc, SIZEPALETTE);  // should be 256

	if (staticColours > 0)
	{
		if (paletteSize != 256)
		{
			ReleaseDC(hwnd,hdc);
			return 1099 ;            // can't support funny hardware palettes!
		}

		GetSystemPaletteEntries ( hdc, 0, paletteSize, sysPal );

		LPRGBQUAD pColours = lpColourTable;

		// add in lower static colours 
		for (i = 0 ; i < staticColours/2 ; i++ )
		{
			pColours->rgbBlue  = sysPal[i].peBlue;
			pColours->rgbRed   = sysPal[i].peRed;
			pColours->rgbGreen = sysPal[i].peGreen;
			pColours->rgbReserved = 0 ;

			pColours++ ;

			// any others?
		}

		// followed by higher static colours
		
		pColours = lpColourTable + paletteSize - staticColours/2;

		for ( i = paletteSize - staticColours/2 ; i < paletteSize ; i++ )
		{
			pColours->rgbBlue  = sysPal[i].peBlue;
			pColours->rgbRed   = sysPal[i].peRed;
			pColours->rgbGreen = sysPal[i].peGreen;
			pColours->rgbReserved = 0 ;

			pColours++ ;
		}

		// set up remaining colours - either from colour table
		// passed in, or just set them to zero

		pColours = lpColourTable + staticColours/2 ;

		if ( colourArray != NULL )
		{
			for ( i = staticColours/2 ; i < 256 - staticColours/2 ; i++ )
			{
				pColours->rgbBlue  = GetBValue(colourArray[i]) ;
				pColours->rgbRed   = GetRValue(colourArray[i]) ; 
				pColours->rgbGreen = GetGValue(colourArray[i]) ;
				pColours->rgbReserved = 0 ;
				pColours++ ;
			}
		}
		else
		{
			for ( i = 0 ; i < 256 - staticColours ; i++ )
			{
				pColours->rgbBlue  = 0 ;
				pColours->rgbRed   = 0 ; 
				pColours->rgbGreen = 0 ;
				pColours->rgbReserved = 0 ;
				pColours++ ;
			}
		}
	}
	else
	{
		// system is set to more than 256 colours. Therefore, there is
		// no system palette, and no static colours. So we'll set them
		// up manually!

		int fc[20][3] ;  // fixed colours

		fc[0][0] = 0x00 ; fc[0][1] = 0x00 ; fc[0][2] = 0x00 ;
		fc[1][0] = 0x80 ; fc[1][1] = 0x00 ; fc[1][2] = 0x00 ;
		fc[2][0] = 0x00 ; fc[2][1] = 0x80 ; fc[2][2] = 0x00 ;
		fc[3][0] = 0x80 ; fc[3][1] = 0x80 ; fc[3][2] = 0x00 ;
		fc[4][0] = 0x00 ; fc[4][1] = 0x00 ; fc[4][2] = 0x80 ;
		fc[5][0] = 0x80 ; fc[5][1] = 0x00 ; fc[5][2] = 0x80 ;
		fc[6][0] = 0x00 ; fc[6][1] = 0x80 ; fc[6][2] = 0x80 ;
		fc[7][0] = 0xC0 ; fc[7][1] = 0xC0 ; fc[7][2] = 0xC0 ;
		fc[8][0] = 0xC0 ; fc[8][1] = 0xDC ; fc[8][2] = 0xC0 ;
		fc[9][0] = 0xA6 ; fc[9][1] = 0xCA ; fc[9][2] = 0xF0 ;
		fc[10][0] = 0xFF ; fc[10][1] = 0xFB ; fc[10][2] = 0xF0 ;
		fc[11][0] = 0xA0 ; fc[11][1] = 0xA0 ; fc[11][2] = 0xA4 ;
		fc[12][0] = 0x80 ; fc[12][1] = 0x80 ; fc[12][2] = 0x80 ;
		fc[13][0] = 0xFF ; fc[13][1] = 0x00 ; fc[13][2] = 0x00 ;
		fc[14][0] = 0x00 ; fc[14][1] = 0xFF ; fc[14][2] = 0x00 ;
		fc[15][0] = 0xFF ; fc[15][1] = 0xFF ; fc[15][2] = 0x00 ;
		fc[16][0] = 0x00 ; fc[16][1] = 0x00 ; fc[16][2] = 0xFF ;
		fc[17][0] = 0xFF ; fc[17][1] = 0x00 ; fc[17][2] = 0xFF ;
		fc[18][0] = 0x00 ; fc[18][1] = 0xFF ; fc[18][2] = 0xFF ;
		fc[19][0] = 0xFF ; fc[19][1] = 0xFF ; fc[19][2] = 0xFF ;

		LPRGBQUAD pColours = lpColourTable;

		for ( i = 0 ; i < 10 ; i++ )
		{
			pColours->rgbRed   = fc[i][0];
			pColours->rgbGreen = fc[i][1];
			pColours->rgbBlue  = fc[i][2];
			pColours->rgbReserved = 0 ;

			pColours++ ;
		}

		pColours = lpColourTable + 256 - 10 ;

		for ( i = 10 ; i < 20 ; i++ )
		{
			pColours->rgbRed   = fc[i][0];
			pColours->rgbGreen = fc[i][1];
			pColours->rgbBlue  = fc[i][2];
			pColours->rgbReserved = 0 ;
			pColours++ ;
		}

		// set up remaining colours - either from colour table
		// passed in, or just set them to zero

		pColours = lpColourTable + 10 ;

		if ( colourArray != NULL )
		{
			for ( i = 10 ; i < 256 - 10 ; i++ )
			{
				pColours->rgbBlue  = GetBValue(colourArray[i]) ;
				pColours->rgbRed   = GetRValue(colourArray[i]) ; 
				pColours->rgbGreen = GetGValue(colourArray[i]) ;
				pColours->rgbReserved = 0 ;
				pColours++ ;
			}
		}
		else
		{
			for ( i = 10 ; i < 256 - 10 ; i++ )
			{
				pColours->rgbBlue  = 0 ;
				pColours->rgbRed   = 0 ; 
				pColours->rgbGreen = 0 ;
				pColours->rgbReserved = 0 ;
				pColours++ ;
			}
		}

		staticColours = 20 ;
		paletteSize   = 256 ;
	}

	ReleaseDC(hwnd,hdc);

	// Now fill in the pixels, with value 0

	BYTE pixel = (BYTE)i ;   
 
	BYTE* pPixel = lpBmpPixels;

	int row ;

	int padding = storageWidth - width ;

	for (row = 0 ; row < height ; row++)
	{
		memset (pPixel, pixel , width) ;
		pPixel += width ;
		memset (pPixel, 0 , padding) ;
		pPixel += padding ;
	} 

	bgColour = 0 ;
	bgPixel  = 0 ;
	bgWord = 0 ;
	bgTriple.rgbtRed = 0 ;
	bgTriple.rgbtGreen = 0 ;
	bgTriple.rgbtBlue = 0 ;

	
	pBmpInfoHdr = lpBmpInfoHdr ;
	pPixels = lpBmpPixels ;
	pColourTable = lpColourTable ;
	numOfColours = 256 ;
	this->width = (short)lpBmpInfoHdr->biWidth ;
	this->height = (short)lpBmpInfoHdr->biHeight ;
	bpp = 8 ;
	isDibSection = false ;
	isPalettized = false ;
	isResource = false ;

    return 0 ;

}   

int CDib::addColours ( CDib * pSourceDib ) 
{

	if ( this->pBmpInfoHdr == NULL ) 
		return 11 ;

	if ( pSourceDib->pBmpInfoHdr == NULL )
		return 12 ;

	LPBITMAPINFOHEADER pDestInfoHdr , pSourceInfoHdr ;

	BITMAPINFO * pSourceInfo ;
	BITMAPINFO * pDestInfo  ;

    LPRGBQUAD pDestColourTable ;
    LPRGBQUAD pSourceColourTable ;

	BYTE*  pSourcePixels ;

	pDestInfoHdr = this->pBmpInfoHdr ;

	pSourceInfoHdr = pSourceDib->pBmpInfoHdr ;

	if (pSourceInfoHdr->biSize != sizeof(BITMAPINFOHEADER))
		// check that it is a Windows DIB file (the CBitmap class
		// does not handle Presentation Manager DIB files or
		// DDB bitmaps
	{
		return 1 ;
	}

	if (pDestInfoHdr->biSize != sizeof(BITMAPINFOHEADER))
		// check that it is a Windows DIB file (the CBitmap class
		// does not handle Presentation Manager DIB files or
		// DDB bitmaps
	{
		return 2 ;
	}

	// For adding colours, both must be 256 colours and not palettized

	int destBpp    = pDestInfoHdr->biBitCount ;
	int sourceBpp  = pSourceInfoHdr->biBitCount ;

	if ( destBpp != 8 )
	{
		return 3 ;
	}

	if ( sourceBpp != 8 )
	{
		return 4 ;
	}

	int destNumColours , sourceNumColours ;

	destNumColours = pDestInfoHdr->biClrUsed ;
	if ( destNumColours == 0 )
		destNumColours = 256 ;

	sourceNumColours = pSourceInfoHdr->biClrUsed ;
	if ( sourceNumColours == 0 )
		sourceNumColours = 256 ;

	if ( destNumColours != 256 ) 
		return 5 ;

	if ( sourceNumColours != 256 )
		return 6 ;

	if ( this->isPalettized != pSourceDib->isPalettized ) 
		return 7 ;

    pDestInfo = (BITMAPINFO *)((char*)pDestInfoHdr);
    pSourceInfo = (BITMAPINFO *)((char*)pSourceInfoHdr);

    pDestColourTable = (LPRGBQUAD)((char*)pDestInfoHdr + sizeof(BITMAPINFOHEADER)) ;
    pSourceColourTable = (LPRGBQUAD)((char*)pSourceInfoHdr + sizeof(BITMAPINFOHEADER)) ;

    pSourcePixels = (BYTE*)pSourceInfoHdr + 
                            sizeof(BITMAPINFOHEADER) + 256*sizeof(RGBQUAD) ;

    int used[256] ;

    int pixVal ;

    int i, j ;

    // clear the "used" array to zero

    for (i = 0 ; i < 256 ; i++)
        used[i] = 0 ;

    int width, height , storageWidth ;

    width = pSourceInfoHdr->biWidth ;

    height = pSourceInfoHdr->biHeight ;

    storageWidth = (width + 3) & ~3 ;

    for ( j = 0 ; j < height ; j++)
    {
        for ( i = 0 ;  i < width ; i++)
        {
            pixVal = *pSourcePixels++ ;
            if (pixVal < 256)
                used[pixVal] = 1 ;
        }

        pSourcePixels += storageWidth - width ;
    }

   // find second zeroed dest colour table entry

   int zeroEntries , nextZero , highestColour ;

   zeroEntries = 0 ; 
   nextZero = 0 ;
   highestColour = 0 ;

   for ( i = 0 ; i < 256  ; i++ )
    {
        if ( pDestColourTable[i].rgbRed == 0 &&
             pDestColourTable[i].rgbGreen == 0 &&
             pDestColourTable[i].rgbBlue == 0
            )
        {

             zeroEntries++ ;
             if (zeroEntries == 2)
                 nextZero = i ;   // this is the first entry we can overwrite
        }
        else
		{
            highestColour = i ;
		}
    }

    // Now go through the source bitmap's used colours, check if already in the
    // dest bitmap's table, and add in if not (and there is enough room)

    for ( i = 0 ; i < 256 ; i++ )
    {
        if ( used[i] != 0 )
        {
            int foundIt = 0 ; 
            for ( j = 0 ; j <= highestColour && foundIt == 0 ; j++)
            {
              if (pDestColourTable[j].rgbRed == pSourceColourTable[i].rgbRed &&
                  pDestColourTable[j].rgbGreen == pSourceColourTable[i].rgbGreen &&
                  pDestColourTable[j].rgbBlue == pSourceColourTable[i].rgbBlue 
                 )
			  {
                  foundIt = 1 ; // got a matching colour
			  }
            }

            if (foundIt == 0)
              {  
                  // need to add the colour into the dest colour table

				if (nextZero == 0)
				  {
					 goto stopLoop ;
				  }
                  else
                  {
                      pDestColourTable[nextZero].rgbRed = pSourceColourTable[i].rgbRed;
                      pDestColourTable[nextZero].rgbBlue = pSourceColourTable[i].rgbBlue;
                      pDestColourTable[nextZero].rgbGreen = pSourceColourTable[i].rgbGreen;

                      if (nextZero > highestColour)
                          highestColour = nextZero;
                        
                      int foundZero =  0;

                      for ( ; nextZero < 255 && foundZero == 0 ; nextZero++)
                        {
                          if (pDestColourTable[nextZero+1].rgbRed == 0 && 
                              pDestColourTable[nextZero+1].rgbBlue == 0 &&
                              pDestColourTable[nextZero+1].rgbGreen == 0
                             )
                              foundZero = 1 ;
                        }
                      if ( foundZero == 0)
                          nextZero = 0 ;
                  }
              }    // end "if foundIt == 0"
        }          // end "if used[i] == 0"
    }              // end "for (i = 0 ..." 

stopLoop :

	return 0 ;

}

int CDib::makePalette ( CPalette * pPal )
{
	// create a palette from a DIB's RGB colour table,

	// Even if palettes won't be selected and realized (i.e. screen colour-depth
	// is 16-bits or 24-bits), it is useful to have a palette when dealing with
	// 256-colour images. This allows conversion of rgb colours to index values
	// (using GetNearestPaletteIndex) without having to keep constructing a 
	// palette from the image. It can also be useful as a basis for synchronizing
	// 256-colour images (palBmpSynchronize).

	// work out number of colours in the RGB table
	
	int numColours;

	numColours = this->pBmpInfoHdr->biClrUsed; // number of colours used

	if (numColours == 0)
	{
		if (this->pBmpInfoHdr->biBitCount != 8) // 256 colour BMP?
			return ERROR_NOT_256_COLOURS ;      // only 256 colours supported for palettes!
		else
			numColours = 256 ; // if biClrUsed is zero, then the
							   // maximum number of colours are defined
							   // in the RGBQUAD table (i.e.256)
	}

	RGBQUAD* pColourTable = (LPRGBQUAD)((char*)(this->pBmpInfoHdr) + sizeof(BITMAPINFOHEADER));
	             // Colour table immediately follows Bit Map Info header

	LOGPALETTE* pLogPal = (LOGPALETTE*) malloc (sizeof(LOGPALETTE) +
		                        numColours*sizeof(PALETTEENTRY));
				     // allocate memory for creating a Logical Palette.

	if (pLogPal == NULL)
		return ERROR_CANNOT_ALLOCATE_PALETTE ; // insufficient memory space for logical palette
		// fill in the logical palette

	pLogPal->palVersion = 0x300 ;  // Windows 3.0+ palette
	pLogPal->palNumEntries = numColours;

	for (int ctr = 0 ; ctr < numColours ; ctr++ )
	{
		pLogPal->palPalEntry[ctr].peRed = pColourTable[ctr].rgbRed;
		pLogPal->palPalEntry[ctr].peGreen = pColourTable[ctr].rgbGreen;
		pLogPal->palPalEntry[ctr].peBlue = pColourTable[ctr].rgbBlue;
		pLogPal->palPalEntry[ctr].peFlags = 0 ;
	}

	BOOL bOk ;

	bOk = pPal->CreatePalette(pLogPal); // create a logical palette
	
	free (pLogPal);
	
	if (! bOk)
		return ERROR_CANNOT_CREATE_PALETTE ; // could not create logical palette

	return 0 ;
}

int CDib::palettize ( CPalette * pPal ) 
{
	// Converts a bitmap from using RGB colours to direct use of palette
	// indices. It will create a 1-to-1 colour table, i.e. pixels contain
	// direct palette values. Pixels will be updated.

	int numColours ;

	BITMAPINFO * pBmpInfo = (BITMAPINFO *)((char*)this->pBmpInfoHdr);

	if ( ! (pBmpInfoHdr->biClrUsed == 256 ||
	       (pBmpInfoHdr->biClrUsed == 0 && 
	        pBmpInfoHdr->biBitCount == 8) )
	   )
	   return 2 ; // not 256 colours!

	numColours = pBmpInfoHdr->biClrUsed ;
	if ( numColours == 0 )
		numColours = 256 ;

	BYTE*  pBmpPixels = (BYTE *)pBmpInfoHdr + sizeof(BITMAPINFOHEADER) +
		                       numColours* sizeof(RGBQUAD);

	// check that bitmap contains 256-colour tables

	// get address of bitmap's colour table

	RGBQUAD* pColourTable = (LPRGBQUAD)((char*)pBmpInfoHdr + sizeof(BITMAPINFOHEADER));
             // Colour table immediately follows Bit Map Info header

	unsigned int map[256] ; // translation table

	for (int i = 0 ; i < 256 ; i++ )
	{
		map[i] = pPal->GetNearestPaletteIndex ( 
					                   RGB(pColourTable->rgbRed,
					                       pColourTable->rgbGreen,
						                   pColourTable->rgbBlue ) ) ;
					   
		pColourTable++ ;
	}

	// check if pixel values were already aligned with the palette
	// (this happens if the colour table exactly matches the palette)

	unsigned int j ;

	int is_1_to_1 = 0 ;

	for ( j = 0 ; j < 256 && map[j] == j ; j++ );

	if ( j != 256)
	{
		// need to change all the pixel values if not 1-to-1
	
		// Work out number of bytes. Each line in the picture is padded
		// out to a 32-bit word boundary.

		int lineWidth = (pBmpInfoHdr->biWidth + 3) & ~3;

		int numBytes = pBmpInfoHdr->biHeight * lineWidth;

		for (j = numBytes ; j > 0 ; j--)
		{
			*pBmpPixels = map[*pBmpPixels] ;
			pBmpPixels++;
		}

	}

	// Colour table needs updating now. When using Palette indices, then simply make
	// it a 1-to-1 colour table. If sticking with RGB colours, then fill it with the
	// loaded palette entries.

	if ( TRUE )  // was .... ->> if ( systemBpp <= 8 ) <<-
	{
		pColourTable = (LPRGBQUAD)((char*)pBmpInfoHdr + sizeof(BITMAPINFOHEADER));

		WORD * p1to1 = (WORD*)pColourTable;

		for ( j = 0 ; j < 256 ; j++ )
			*p1to1++ = j ;
	}
	else
	{
		// get a copy of the loaded palette's colours, and copy into the image's
		// colour table

		PALETTEENTRY pal[256];
    
		int numRetrieved;

		numRetrieved = pPal->GetPaletteEntries(0, 256, pal);
	
		if (numRetrieved != 256)
			return 3 ; 
	
		pColourTable = (LPRGBQUAD)((char*)pBmpInfoHdr + sizeof(BITMAPINFOHEADER));
		                      // point at bitmap's colour table

		for (j = 0 ; j < 256 ; j++ )
		{
			pColourTable->rgbRed = pal[j].peRed;
			pColourTable->rgbGreen = pal[j].peGreen;
			pColourTable->rgbBlue = pal[j].peBlue;

			pColourTable++;
		}
	}
	
	// all done!

	this->isPalettized = true ;

	return 0 ;
}

int CDib::makeDibSection ( int iWidth , int iHeight , int iBpp ,
						   CDC * pDC ,
						   CPalette * pPal , bool palettizeIt) 
{
	// work out how many bytes are required for the DIB section.
	//
	// The area for pixels is automatically allocated by Windows.
	//
	// We need to allocate space for the BITMAPINFOHDR and the Colour Table.
	//
	int bytesRequired ;

	bytesRequired = sizeof(BITMAPINFOHEADER) ; // the 3 words are for our header

	if ( iBpp == 8 )
	{
		bytesRequired += (palettizeIt) ? 256*sizeof(WORD) : 256*sizeof(RGBQUAD) ;
	}

	BYTE * pSpace ;

	pSpace = ( BYTE *)malloc ( bytesRequired ) ;

	if ( pSpace == NULL )
	{
		return 1 ;
	}

	memset ( pSpace , '\0' , bytesRequired ) ;

    // set up DIB section's BITMAPINFO 

    LPBITMAPINFOHEADER  pCTInfoHdr ;
	
	pCTInfoHdr = (LPBITMAPINFOHEADER)pSpace ;

    pCTInfoHdr->biSize = sizeof(BITMAPINFOHEADER);
    pCTInfoHdr->biWidth = iWidth ;
    pCTInfoHdr->biHeight = iHeight ;
    pCTInfoHdr->biPlanes = 1 ;
    pCTInfoHdr->biBitCount = iBpp ;
    pCTInfoHdr->biCompression = BI_RGB;
    pCTInfoHdr->biSizeImage = 0 ;
    pCTInfoHdr->biXPelsPerMeter = 0 ;
    pCTInfoHdr->biYPelsPerMeter = 0 ;
    pCTInfoHdr->biClrUsed = 0 ;
    pCTInfoHdr->biClrImportant = 0 ;

    // set up colour table, if it is a 256-colour DIB section

	if (iBpp == 8 )
	{
		if ( palettizeIt ) 
		{
			WORD* pColours = (WORD*)((char*)pCTInfoHdr + sizeof(BITMAPINFOHEADER));

			for (int colourNum = 0 ; colourNum < 256 ; colourNum++)
				*pColours++ = (WORD)colourNum;
		}
		else
		{
			// set up DIB section's colour table based on the palette
			// passed to the routine.

			LPRGBQUAD pColours ;
			
			pColours = (LPRGBQUAD)((char*)pCTInfoHdr + sizeof(BITMAPINFOHEADER));

			PALETTEENTRY pal[256];
    
			int numRetrieved;

			numRetrieved = pPal->GetPaletteEntries(0, 256, pal);
	
			if (numRetrieved != 256)
			{
				free ( pSpace ) ;
					return 2 ;
			}
	
			for (int j = 0 ; j < 256 ; j++ )
			{
				pColours->rgbRed = pal[j].peRed;
				pColours->rgbGreen = pal[j].peGreen;
				pColours->rgbBlue = pal[j].peBlue;
				pColours++;
			}
		}
	}

    // now create the DIB section!!

	if ( palettizeIt )
	{
		(void)pDC->SelectPalette ( pPal, FALSE ) ; // select into foreground

		(void)pDC->RealizePalette () ;
	}

	HBITMAP dibSect ;
	BYTE *  pDibPixels ;

    dibSect = CreateDIBSection ( pDC->m_hDC ,
		                         (LPBITMAPINFO)pSpace ,
                                 (palettizeIt) ? DIB_PAL_COLORS : DIB_RGB_COLORS,
                                 (VOID**)&pDibPixels,
                                 NULL,
                                 0
                               );

	if (dibSect == NULL)
	{
		free ( pSpace ) ;
		return 4 ;
	}

	this->pBmpInfoHdr  = (LPBITMAPINFOHEADER)pSpace ;
	this->pPixels      = pDibPixels ;
	this->pColourTable = (RGBQUAD*)(pSpace + sizeof(BITMAPINFOHEADER)) ;
	this->numOfColours = (iBpp == 8) ? 256 : 0 ;
	this->width        = iWidth ;
	this->height       = iHeight ;
	this->bpp          = iBpp ;
	this->isDibSection = true ;
	this->hBmp         = dibSect ;
	this->isPalettized = palettizeIt ;

	return 0 ;
}

void CDib::backcpy ( char * pDest , char * pSource , int num )
{
	// same as memcpy, but copies backwards

	pSource += num - 1 ;

	for ( int i = 0 ; i < num ; i++ ) 
		*pDest++ = *pSource-- ;
}

int CDib::solidMerge ( int  destXpos , int destYpos ,
		               LPBITMAPINFOHEADER psourceBmpInfoHdr ,
					   BYTE * psourceBmpPixels ,
					   int sourceXpos , int sourceYpos ,
					   int width ,
					   int height ,
					   bool reversed )
{
	LPBITMAPINFOHEADER pdestBmpInfoHdr = this->pBmpInfoHdr ;

	BYTE * pdestBmpPixels = this->pPixels ;

	int	sourceBpp = psourceBmpInfoHdr->biBitCount;
	int   destBpp = pdestBmpInfoHdr->biBitCount ;

	LPBITMAPINFO pdestBmpInfo = (BITMAPINFO *)((char*)pdestBmpInfoHdr);
	LPBITMAPINFO psourceBmpInfo  = (BITMAPINFO *)((char*)psourceBmpInfoHdr);

	// If sourceX or sourceY is negative, or destX or destY is negative, adjust
	// position and height and width appropriately so that they are all >= 0 !!

	if ( sourceXpos < 0 )
	{
		destXpos -= sourceXpos ;
		width    += sourceXpos ;
		sourceXpos = 0 ;
	}
	if ( sourceYpos < 0 )
	{
		destYpos -= sourceYpos ;
		height   += sourceYpos ;
		sourceYpos = 0 ;
	}

	if ( destXpos < 0 )
	{
		sourceXpos -= destXpos ;
		width      += destXpos ;
		destXpos   = 0 ;
	}
	if ( destYpos < 0 )
	{
		sourceYpos -= destYpos ;
		height     += destYpos ;
		destYpos    = 0 ;
	}

	if ( width < 0 || height < 0 )
		return 1 ;

	if ( destXpos >= pdestBmpInfoHdr->biWidth || destYpos >= pdestBmpInfoHdr->biHeight )
	{
		return 1 ;
	}

	// same check for source image

	if ( sourceXpos >= psourceBmpInfoHdr->biWidth || 
		 sourceYpos >= psourceBmpInfoHdr->biHeight )
	{
		return 2 ;
	}

	if (width == 0)
		width = psourceBmpInfoHdr->biWidth;

	if (height == 0)
		height = psourceBmpInfoHdr->biHeight;

	// Make sure not copying beyond the bounds of the input image

	if ( sourceXpos + width >= psourceBmpInfoHdr->biWidth )
		width = psourceBmpInfoHdr->biWidth - sourceXpos ;

	if ( sourceYpos + height >= psourceBmpInfoHdr->biHeight )
		height = psourceBmpInfoHdr->biHeight - sourceYpos ;

	if (destXpos + width >= pdestBmpInfoHdr->biWidth)
		width = pdestBmpInfoHdr->biWidth - destXpos ;

	if (destYpos + height >= pdestBmpInfoHdr->biHeight)
		height = pdestBmpInfoHdr->biHeight - destYpos  ;

		// work out actual width of each line in bitmap (each line gets padded to a
	// 32-bit word boundary)

	// 16 bpp takes 2 bytes per colour, 24 bpp takes 3 bytes

	// work out width of each line in bytes

	int destWidth , destHeight , sourceWidth , sourceHeight ;

	int destByteWidth , sourceByteWidth ;

	int destStoreWidth , destStoreHeight , sourceStoreWidth , sourceStoreHeight ;

	int destPad , sourcePad ;

	BYTE * pDest , * pSource ;

	destWidth = pdestBmpInfoHdr->biWidth ;
	destHeight = pdestBmpInfoHdr->biHeight ;
	sourceWidth = psourceBmpInfoHdr->biWidth ;
	sourceHeight = psourceBmpInfoHdr->biHeight ;

	destByteWidth = (pdestBmpInfoHdr->biWidth * destBpp/8 ) ;
	sourceByteWidth = (psourceBmpInfoHdr->biWidth * sourceBpp/8 ) ;

	destStoreWidth = (destByteWidth + 3 ) & ~3 ;
	sourceStoreWidth = (sourceByteWidth + 3 ) & ~3 ;

	destStoreHeight = pdestBmpInfoHdr->biHeight ;
	sourceStoreHeight = psourceBmpInfoHdr->biHeight;

	// work out number of extra bytes at end of each stored line

	destPad = destStoreWidth - width*destBpp/8 ;
	sourcePad = sourceStoreWidth - width*sourceBpp/8 ;

	pDest = pdestBmpPixels + destStoreWidth*(destStoreHeight - destYpos - height )
		             + destXpos*destBpp/8;

	pSource = psourceBmpPixels + sourceStoreWidth*(sourceStoreHeight - sourceYpos - height ) 
					 + sourceXpos*sourceBpp/8 ;

	int line ;

	if ( destBpp == 8 )
	{
		if ( sourceBpp == 8 )
		{
			// merging 256-colour image into 256-colour image

			// direct byte to byte copying

			for (line = 0 ; line < height ; line++)
			{
				if ( !reversed )
					memcpy (pDest, pSource, width);
				else
					backcpy ( (char*)pDest , (char*)pSource , width);
				pDest += destStoreWidth;
				pSource += sourceStoreWidth;
			}
		}
		else
		{
			return 3 ;    // can't copy 16 bpp or 24 bpp into 8 bpp
		}
	}
	else if ( destBpp == 16 )
	{
		// copying into a 16-bit image

		if ( sourceBpp == 16 )
		{
			// copying with same colour-depth

			for (line = 0 ; line < height ; line++)
			{
				if ( !reversed)
					memcpy (pDest, pSource, width*2);
				else
					backcpy ( (char*)pDest, (char*)pSource, width*2) ;
				pDest += destStoreWidth;
				pSource += sourceStoreWidth;
			}
		}
		else if (sourceBpp == 24 ) 
		{
			// copying 24-bit image into 16-bit
			// Need to reduce each pixel from 3 bytes down to 2

			RGBTRIPLE * pS = (RGBTRIPLE*)pSource ;
			WORD      * pD = (WORD *)pDest ; 

			int col ;

			for (line = 0 ; line < height ; line++)
			{
				if (!reversed)
				{
					for ( col = 0 ; col < width ; col++ )
					{
						*pD++ =  ((pS->rgbtRed>>3)<<10) +
							     ((pS->rgbtGreen>>3)<<5) +
								  (pS->rgbtBlue>>3) ;
						pS++ ;
					}
				}
				else
				{
					pS += width ;
					RGBTRIPLE * pEnd = pS ;

					for (col = 0 ; col < width ; col++)
					{
						pEnd-- ;
						*pD++ =  ((pEnd->rgbtRed>>3)<<10) +
							     ((pEnd->rgbtGreen>>3)<<5) +
								  (pEnd->rgbtBlue>>3) ;
					}
				}
					
				// advance pointers to next line

				pD += destPad/2;
				pS = (RGBTRIPLE*)((char*)pS + sourcePad) ; 
				           // need to do it this way, because there are 3 bytes 
				           // for each pixel, and the byte length of each line may
				           // not be exactly divisible by three.
			}
		}
		else if ( sourceBpp == 8 )
		{
			// copying 256-colour image into 16-bit image

			// not allowed if source image is palettized ... hopefully, that will
			// already have been checked

			WORD      * pD = (WORD *)pDest ; 

			// need to access the colour table to work out each rgb colour

			RGBQUAD* pColourTable = (LPRGBQUAD)((char*)psourceBmpInfoHdr 
				                          + sizeof(BITMAPINFOHEADER));
	        
			int col , ctNum ;

			for (line = 0 ; line < height ; line++)
			{
				if (!reversed)
				{
					for ( col = 0 ; col < width ; col++ )
					{
						ctNum = *pSource++ ; // colour table index
						*pD++ = (((pColourTable[ctNum].rgbRed)>>3)<<10) +
						        (((pColourTable[ctNum].rgbGreen)>>3)<<5) +
						  	    ((pColourTable[ctNum].rgbBlue)>>3) ;
					}
				}
				else
				{
					pSource += width ;
					BYTE * pEnd = pSource ;

					for (col = 0 ; col < width ; col++)
					{
						ctNum = *(--pEnd);
						*pD++ = (((pColourTable[ctNum].rgbRed)>>3)<<10) +
						        (((pColourTable[ctNum].rgbGreen)>>3)<<5) +
						  	    ((pColourTable[ctNum].rgbBlue)>>3) ;
					}
				}

					
				// advance pointers to next line

				pD += destPad/2;
				pSource += sourcePad ;
			}
		}
		else
		{
			return 6 ;
		}
	}
	else if ( destBpp == 24 )
	{
		// copying into a 24-bit image

		if ( sourceBpp == 24 )
		{
			// copying with same colour-depth

			for (line = 0 ; line < height ; line++)
			{
				if (!reversed)
					memcpy (pDest, pSource, width*3);
				else
					backcpy ( (char*)pDest , (char*)pSource , width*3 ) ;
				pDest += destStoreWidth;
				pSource += sourceStoreWidth;
			}
		}
		else if (sourceBpp == 16 ) 
		{
			// copying 16-bit image into 24-bit
			// Need to increase each pixel from 2 bytes to 3

			RGBTRIPLE * pD = (RGBTRIPLE*)pDest ;
			WORD      * pS = (WORD *)pSource ; 

			WORD pixel ;

			int col ;

			for (line = 0 ; line < height ; line++)
			{
				if (!reversed)
				{
					for ( col = 0 ; col < width ; col++ )
					{
						pixel = *pS++ ;
						pD->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
						pD->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
						pD->rgbtBlue = (pixel & 0x001F) << 3 ;
						pD++ ;
					}
				}
				else
				{
					pS += width ;
					WORD * pEnd = pS ;
					for ( col = 0 ; col < width ; col++ )
					{
						pixel = *(--pEnd) ;
						pD->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
						pD->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
						pD->rgbtBlue = (pixel & 0x001F) << 3 ;
						pD++ ;
					}
				}
					
				// advance pointers to next line

				pD = (RGBTRIPLE*)((char*)pD + destPad);
				pS += sourcePad / 2 ;
			}
		}
		else if ( sourceBpp == 8 )
		{
			// copying 256-colour image into 24-bit image

			// not allowed if source image is palettized ... hopefully, that will
			// already have been checked

			RGBTRIPLE      * pD = (RGBTRIPLE *)pDest ; 

			// need to access the colour table to work out each rgb colour

			RGBQUAD* pColourTable = (LPRGBQUAD)((char*)psourceBmpInfoHdr 
				                          + sizeof(BITMAPINFOHEADER));
	        
			int col , ctNum ;

			for (line = 0 ; line < height ; line++)
			{
				if ( !reversed)
				{
					for ( col = 0 ; col < width ; col++ )
					{
						ctNum = *pSource++ ; // colour table index
						pD->rgbtBlue = pColourTable[ctNum].rgbBlue ;
						pD->rgbtGreen = pColourTable[ctNum].rgbGreen ;
						pD->rgbtRed = pColourTable[ctNum].rgbRed ;

						pD++ ;
					}
				}
				else
				{
					pSource += width ;
					BYTE * pEnd = pSource ;

					for ( col = 0 ; col < width ; col++ )
					{
						ctNum = *(--pEnd) ;
						pD->rgbtBlue = pColourTable[ctNum].rgbBlue ;
						pD->rgbtGreen = pColourTable[ctNum].rgbGreen ;
						pD->rgbtRed = pColourTable[ctNum].rgbRed ;

						pD++ ;
					}
				}
					
				// advance pointers to next line
				pD = (RGBTRIPLE*)((char*)pD + destPad);
				pSource += sourcePad ;
			}
		}
		else
		{
			return 7 ;
		}
	}
	else
	{
		return 9 ;
	}
			
    return 0 ;
}

int CDib::transpMerge ( int  destXpos , int destYpos ,
		                LPBITMAPINFOHEADER psourceBmpInfoHdr ,
					    BYTE * psourceBmpPixels ,
					    int sourceXpos , int sourceYpos ,
					    int width ,
					    int height ,
					    bool reversed ,
					    unsigned int  transpPixel , bool inverted ) 
{
	// Does a transparent copy.
	//
	// 'transpPixel' is the pixel value of the transparent colour.
	// For 8 bpp images, this will be a palette index.
	// For 16 and 24 bpp images, this will be a colour
	//
	// 'reversed' means the image will be copied right-to-left, i.e.
	// mirror image.
	//
	// 'inverted' means that pixels are only copied if they are the SAME
	// as the transparent colour. This allows dynamic mask creation.

	LPBITMAPINFOHEADER pdestBmpInfoHdr = this->pBmpInfoHdr ;

	BYTE * pdestBmpPixels = this->pPixels ;

	int	sourceBpp = psourceBmpInfoHdr->biBitCount;
	int   destBpp = pdestBmpInfoHdr->biBitCount ;

	LPBITMAPINFO pdestBmpInfo = (BITMAPINFO *)((char*)pdestBmpInfoHdr);
	LPBITMAPINFO psourceBmpInfo  = (BITMAPINFO *)((char*)psourceBmpInfoHdr);

		if ( sourceXpos < 0 )
	{
		destXpos -= sourceXpos ;
		width    += sourceXpos ;
		sourceXpos = 0 ;
	}
	if ( sourceYpos < 0 )
	{
		destYpos -= sourceYpos ;
		height   += sourceYpos ;
		sourceYpos = 0 ;
	}

	if ( destXpos < 0 )
	{
		sourceXpos -= destXpos ;
		width      += destXpos ;
		destXpos   = 0 ;
	}

	if ( destYpos < 0 )
	{
		sourceYpos -= destYpos ;
		height     += destYpos ;
		destYpos    = 0 ;
	}

	if ( width < 0 || height < 0 )
		return 1 ;

	if ( destXpos >= pdestBmpInfoHdr->biWidth || destYpos >= pdestBmpInfoHdr->biHeight )
	{
		return 1 ;
	}

	// same check for source image

	if ( sourceXpos >= psourceBmpInfoHdr->biWidth || 
		 sourceYpos >= psourceBmpInfoHdr->biHeight )
	{
		return 2 ;
	}

	if (width == 0)
		width = psourceBmpInfoHdr->biWidth;

	if (height == 0)
		height = psourceBmpInfoHdr->biHeight;

	// Make sure not copying beyond the bounds of the input image

	if ( sourceXpos + width >= psourceBmpInfoHdr->biWidth )
		width = psourceBmpInfoHdr->biWidth - sourceXpos ;

	if ( sourceYpos + height >= psourceBmpInfoHdr->biHeight )
		height = psourceBmpInfoHdr->biHeight - sourceYpos ;

	if (destXpos + width >= pdestBmpInfoHdr->biWidth)
		width = pdestBmpInfoHdr->biWidth - destXpos ;

	if (destYpos + height >= pdestBmpInfoHdr->biHeight)
		height = pdestBmpInfoHdr->biHeight - destYpos  ;

		// work out actual width of each line in bitmap (each line gets padded to a
	// 32-bit word boundary)

	// 16 bpp takes 2 bytes per colour, 24 bpp takes 3 bytes

	// work out width of each line in bytes

	int destWidth , destHeight , sourceWidth , sourceHeight ;

	int destByteWidth , sourceByteWidth ;

	int destStoreWidth , destStoreHeight , sourceStoreWidth , sourceStoreHeight ;

	int destPad , sourcePad ;

	BYTE * pDest , * pSource ;

	destWidth = pdestBmpInfoHdr->biWidth ;
	destHeight = pdestBmpInfoHdr->biHeight ;
	sourceWidth = psourceBmpInfoHdr->biWidth ;
	sourceHeight = psourceBmpInfoHdr->biHeight ;

	destByteWidth = (pdestBmpInfoHdr->biWidth * destBpp/8 ) ;
	sourceByteWidth = (psourceBmpInfoHdr->biWidth * sourceBpp/8 ) ;

	destStoreWidth = (destByteWidth + 3 ) & ~3 ;
	sourceStoreWidth = (sourceByteWidth + 3 ) & ~3 ;

	destStoreHeight = pdestBmpInfoHdr->biHeight ;
	sourceStoreHeight = psourceBmpInfoHdr->biHeight;

	// work out number of extra bytes at end of each stored line

	destPad = destStoreWidth - width*destBpp/8 ;
	sourcePad = sourceStoreWidth - width*sourceBpp/8 ;

	pDest = pdestBmpPixels + destStoreWidth*(destStoreHeight - destYpos - height )
		             + destXpos*destBpp/8;

	pSource = psourceBmpPixels + sourceStoreWidth*(sourceStoreHeight - sourceYpos - height ) 
					 + sourceXpos*sourceBpp/8 ;

	if ( destBpp == 8 )
	{
		if ( sourceBpp == 8 )
		{
			// merging 256-colour image into 256-colour image

			// transparent copying of 256 colour image

			BYTE pixel , transpByte ;

			BYTE * pD ;

			int row , col ;

			transpByte = (BYTE)transpPixel ;

			if ( ! reversed ) 
			{
				if ( ! inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel != transpByte)
								*(pDest++) = pixel;
							else
								pDest++;
						}
						// next line
						pSource += sourcePad; // = sourceStoreWidth - width
						pDest   += destPad;   // = destStoreWidth - width
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel == transpByte)
								*pDest++ = pixel;
							else
								pDest++;
						}
						// next line
						pSource += sourcePad;
						pDest   += destPad;
					}
				}
			}
			else
			{
				if ( ! inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						pD = pDest + destByteWidth ;

						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel != transpByte)
								*(--pD) = pixel;
							else
								--pD ;
						}
						// next line
						pSource += sourcePad; // = sourceStoreWidth - width
						pDest   += destStoreWidth ;   // = destStoreWidth - width
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pD = pDest + destByteWidth ;

						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel == transpByte)
								*(--pD) = pixel;
							else
								--pD;
						}
						// next line
						pSource += sourcePad;
						pDest   += destStoreWidth ;
					}
				}
			}

		}
		else
		{
			return 3 ;    // can't copy 16 bpp or 24 bpp into 8 bpp
		}
	}
	else if ( destBpp == 16 )
	{
		// copying into a 16-bit image

		if ( sourceBpp == 16 )
		{
			// copying with same colour-depth

			// transparent copying of 16-bit image to a 16-bit image

			WORD pixel ;

			int row , col ;

			WORD * pS = (WORD*)pSource ;
			WORD * pD = (WORD*)pDest ;

			WORD * pDx ;

			if ( ! reversed )
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel != transpPixel)
								*pD++ = pixel;
							else
								pD++;
						}
						// next line

						pS += sourcePad/2;  // we can get away with this because we know
						                    // that 2 pixels exactly fit into each word,
						                    // and there are two bytes per pixel. The 
						                    // byte length of each line is always a multiple
						                    // of 4, so everything is evenly divisible.
						pD += destPad/2;
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel == transpPixel)
								*pD++ = pixel;
							else
								pD++;
						}
						// next line
						pS += sourcePad/2;
						pD += destPad/2;
					}
				}
			}
			else
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + (destByteWidth/2) ;

						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel != transpPixel)
								*(--pDx) = pixel;
							else
								--pDx ;
						}
						// next line

						pS += sourcePad/2;  // we can get away with this because we know
						                    // that 2 pixels exactly fit into each word,
						                    // and there are two bytes per pixel. The 
						                    // byte length of each line is always a multiple
						                    // of 4, so everything is evenly divisible.
						pD += destStoreWidth/2;
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + (destByteWidth/2);

						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel == transpPixel)
								*(--pDx) = pixel;
							else
								--pDx;
						}
						// next line
						pS += sourcePad/2;
						pD += destStoreWidth/2;
					}
				}
			}

		}
		else if (sourceBpp == 24 ) 
		{
			// copying 24-bit image into 16-bit
			// Need to reduce each pixel from 3 bytes down to 2

			// transparent copying of 24-bit image to a 16-bit image

			RGBTRIPLE pixel , transpTriple ;

			RGBTRIPLE * pS = (RGBTRIPLE*)pSource ;
			WORD      * pD = (WORD *)pDest ; 

			int row , col ;

			transpTriple = *(RGBTRIPLE*)(&transpPixel) ;

			WORD * pDx ;

			if ( ! reversed ) 
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
						
							if (pixel.rgbtBlue != transpTriple.rgbtBlue ||
							 	pixel.rgbtGreen != transpTriple.rgbtGreen ||
							 	pixel.rgbtRed != transpTriple.rgbtRed)
	
								*pD++ = ((pixel.rgbtRed>>3)<<10) +
								        ((pixel.rgbtGreen>>3)<<5) +
										 (pixel.rgbtBlue>>3) ;
							else
								pD++;
						}
						// next line
	
						pD += destPad/2;
					    pS = (RGBTRIPLE*)((char*)pS + sourcePad) ; 
				
					}

				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel .rgbtBlue == transpTriple.rgbtBlue &&
							 	pixel.rgbtGreen == transpTriple.rgbtGreen &&
							 	pixel.rgbtRed == transpTriple.rgbtRed)
	
								*pD++ = ((pixel.rgbtRed>>3)<<10) +
								        ((pixel.rgbtGreen>>3)<<5)  +
										(pixel.rgbtBlue>>3) ;
							else
								pD++;
						}
						// next line	

						pD += destPad/2;
						pS = (RGBTRIPLE*)((char*)pS + sourcePad) ; 
						
					}
				}
			}
			else
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + (destByteWidth/2) ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
						
							if (pixel.rgbtBlue != transpTriple.rgbtBlue ||
							 	pixel.rgbtGreen != transpTriple.rgbtGreen ||
							 	pixel.rgbtRed != transpTriple.rgbtRed)
	
								*(--pDx) = ((pixel.rgbtRed>>3)<<10) +
								          ((pixel.rgbtGreen>>3)<<5) +
										  (pixel.rgbtBlue>>3) ;
							else
								--pDx ;
						}
						// next line
	
						pD += destStoreWidth/2;
					    pS = (RGBTRIPLE*)((char*)pS + sourcePad) ; 
				
					}

				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + (destByteWidth/2) ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel.rgbtBlue == transpTriple.rgbtBlue &&
							 	pixel.rgbtGreen == transpTriple.rgbtGreen &&
							 	pixel.rgbtRed == transpTriple.rgbtRed)
	
								*(--pDx) = ((pixel.rgbtRed>>3)<<10) +
								           ((pixel.rgbtGreen>>3)<<5)  +
										   (pixel.rgbtBlue>>3) ;
							else
								--pDx;
						}
						// next line	

						pD += destStoreWidth/2;
						pS = (RGBTRIPLE*)((char*)pS + sourcePad) ; 
						
					}
				}
			}

		}
		else if ( sourceBpp == 8 )
		{
			// copying 256-colour image into 16-bit image

			// not allowed if source image is palettized ... hopefully, that will
			// already have been checked

			// transparent copying of 8-bit image to a 16-bit image

			BYTE pixel , transpByte ;

			WORD      * pD = (WORD *)pDest ; 

			WORD      * pDx ;

			RGBQUAD* pColourTable = (LPRGBQUAD)((char*)psourceBmpInfoHdr 
				                          + sizeof(BITMAPINFOHEADER));

			int row , col ;

			transpByte = (BYTE)transpPixel ;

			if ( ! reversed ) 
			{
				if ( ! inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel != transpByte)
	
								*pD++ = (((pColourTable[pixel].rgbRed)>>3)<<10) +
									    (((pColourTable[pixel].rgbGreen)>>3)<<5) +
									    ((pColourTable[pixel].rgbBlue)>>3) ;
							else
								pD++;
						}
						// next line
	
						pSource += sourcePad ;
						pD += destPad/2 ;
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel == transpByte)
								*pD++ = (((pColourTable[pixel].rgbRed)>>3)<<10) +
									    (((pColourTable[pixel].rgbGreen)>>3)<<5) +
									    ((pColourTable[pixel].rgbBlue)>>3) ;
							else
								pD++;
						}
						// next line
	
						pSource += sourcePad ;
						pD += destPad/2 ;
					}
				}
			}
			else
			{
				if ( ! inverted )
				{
					pDx = pD + destByteWidth/2 ;
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel != transpByte)
	
								*(--pDx) = (((pColourTable[pixel].rgbRed)>>3)<<10) +
									    (((pColourTable[pixel].rgbGreen)>>3)<<5) +
									    ((pColourTable[pixel].rgbBlue)>>3) ;
							else
								--pDx ;
						}
						// next line
	
						pSource += sourcePad ;
						pD += destStoreWidth/2 ;
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/2 ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel == transpByte)
								*(--pDx) = (((pColourTable[pixel].rgbRed)>>3)<<10) +
									    (((pColourTable[pixel].rgbGreen)>>3)<<5) +
									    ((pColourTable[pixel].rgbBlue)>>3) ;
							else
								--pDx ;
						}
						// next line
	
						pSource += sourcePad ;
						pD += destStoreWidth/2 ;
					}
				}
			}
		}
		else
		{
			return 6 ;
		}
	}
	else if ( destBpp == 24 )
	{
		// copying into a 24-bit image

		if ( sourceBpp == 24 )
		{
			// copying with same colour-depth

			// transparent copying of 24-bit image to a 24-bit image

			RGBTRIPLE pixel , transpTriple ;

			int row , col ;

			RGBTRIPLE * pS = (RGBTRIPLE*)pSource ;
			RGBTRIPLE * pD = (RGBTRIPLE*)pDest ;

			RGBTRIPLE * pDx ;

			transpTriple = *(RGBTRIPLE*)(&transpPixel) ;

			if ( ! reversed ) 
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel.rgbtBlue != transpTriple.rgbtBlue ||
							 	pixel.rgbtGreen != transpTriple.rgbtGreen ||
							 	pixel.rgbtRed != transpTriple.rgbtRed)
								*pD++ = pixel;
							else
								pD++;
						}
						// next line
	
						pS = (RGBTRIPLE*)((char*)pS + sourcePad);
						pD = (RGBTRIPLE*)((char*)pD + destPad);
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel.rgbtBlue == transpTriple.rgbtBlue &&
							 	pixel.rgbtGreen == transpTriple.rgbtGreen &&
							 	pixel.rgbtRed == transpTriple.rgbtRed)
								*pD++ = pixel;
							else
								pD++;
						}
						// next line
	
						pS = (RGBTRIPLE*)((char*)pS + sourcePad);
						pD = (RGBTRIPLE*)((char*)pD + destPad);
					}
				}
			}
			else
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/3 ;

						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel.rgbtBlue != transpTriple.rgbtBlue ||
							 	pixel.rgbtGreen != transpTriple.rgbtGreen ||
							 	pixel.rgbtRed != transpTriple.rgbtRed)
								*(--pDx) = pixel;
							else
								--pDx ;
						}
						// next line
	
						pS = (RGBTRIPLE*)((char*)pS + sourcePad);
						pD = (RGBTRIPLE*)((char*)pD + destStoreWidth);
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/3 ;

						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel.rgbtBlue == transpTriple.rgbtBlue &&
							 	pixel.rgbtGreen == transpTriple.rgbtGreen &&
							 	pixel.rgbtRed == transpTriple.rgbtRed)
								*(--pDx) = pixel;
							else
								--pDx;
						}
						// next line
	
						pS = (RGBTRIPLE*)((char*)pS + sourcePad);
						pD = (RGBTRIPLE*)((char*)pD + destStoreWidth);
					}
				}
			}
		}
		else if (sourceBpp == 16 ) 
		{
			// copying 16-bit image into 24-bit
			// Need to increase each pixel from 2 bytes to 3

			// transparent copying of 16-bit image to a 24-bit image

			WORD pixel ;

			RGBTRIPLE * pD = (RGBTRIPLE*)pDest ;
			WORD      * pS = (WORD *)pSource ; 

			RGBTRIPLE * pDx ;

			int row , col ;

			if ( ! reversed ) 
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if ( pixel != transpPixel )  
							{
								pD->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
								pD->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
								pD->rgbtBlue = (pixel & 0x001F) << 3 ;
								pD++ ;
							}
							else
								pD++ ;
						}
						// next line
	
						pS += sourcePad/2 ;
						pD = (RGBTRIPLE*)((char*)pD + destPad);
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel == transpPixel)
							{
								pD->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
								pD->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
								pD->rgbtBlue = (pixel & 0x001F) << 3 ;
								pD++ ;
							}
								
							else
								pD++;
						}
						// next line
	
						pS += sourcePad/2;
						pD = (RGBTRIPLE*)((char*)pD + destPad);
					}
				}
			}
			else
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/3 ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if ( pixel != transpPixel )  
							{
								--pDx ;
								pDx->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
								pDx->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
								pDx->rgbtBlue = (pixel & 0x001F) << 3 ;
							}
							else
								--pDx ;
						}
						// next line
	
						pS += sourcePad/2 ;
						pD = (RGBTRIPLE*)((char*)pD + destStoreWidth);
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/3 ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pS++;
							if (pixel == transpPixel)
							{
								--pDx ;
								pDx->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
								pDx->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
								pDx->rgbtBlue = (pixel & 0x001F) << 3 ;
							}
							else
								--pDx;
						}
						// next line
	
						pS += sourcePad/2;
						pD = (RGBTRIPLE*)((char*)pD + destStoreWidth);
					}
				}
			}

		}
		else if ( sourceBpp == 8 )
		{
			// copying 256-colour image into 24-bit image

			// not allowed if source image is palettized ... hopefully, that will
			// already have been checked

			BYTE pixel , transpByte ;

			RGBTRIPLE      * pD = (RGBTRIPLE *)pDest ; 

			RGBTRIPLE * pDx ;

			RGBQUAD* pColourTable = (LPRGBQUAD)((char*)psourceBmpInfoHdr 
				                          + sizeof(BITMAPINFOHEADER));

			int row , col ;

			transpByte = (BYTE)transpPixel ;

			if ( ! reversed ) 
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if ( pixel != transpByte )  
							{
								pD->rgbtBlue = pColourTable[pixel].rgbBlue ;
								pD->rgbtGreen = pColourTable[pixel].rgbGreen ;
								pD->rgbtRed = pColourTable[pixel].rgbRed ;
								pD++ ;
							}
							else
								pD++ ;
						}
						// next line	

						pSource += sourcePad ;
						pD = (RGBTRIPLE*)((char*)pD + destPad);
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel == transpByte)
							{
								pD->rgbtBlue = pColourTable[pixel].rgbBlue ;
								pD->rgbtGreen = pColourTable[pixel].rgbGreen ;
								pD->rgbtRed = pColourTable[pixel].rgbRed ;

								pD++ ;
							}
							else
								pD++;
						}
						// next line	

						pSource += sourcePad;
						pD = (RGBTRIPLE*)((char*)pD + destPad);
					}
				}
			}
			else
			{
				if ( !inverted )
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/3 ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if ( pixel != transpByte )  
							{
								--pDx ;
								pDx->rgbtBlue = pColourTable[pixel].rgbBlue ;
								pDx->rgbtGreen = pColourTable[pixel].rgbGreen ;
								pDx->rgbtRed = pColourTable[pixel].rgbRed ;
							}
							else
								--pDx ;
						}
						// next line	

						pSource += sourcePad ;
						pD = (RGBTRIPLE*)((char*)pD + destStoreWidth);
					}
				}
				else
				{
					for (row = 0 ; row < height ; row++)
					{
						pDx = pD + destByteWidth/3 ;
						for (col = 0 ; col < width ; col++)
						{
							pixel = *pSource++;
							if (pixel == transpByte)
							{
								--pDx ;
								pDx->rgbtBlue = pColourTable[pixel].rgbBlue ;
								pDx->rgbtGreen = pColourTable[pixel].rgbGreen ;
								pDx->rgbtRed = pColourTable[pixel].rgbRed ;
							}
							else
								--pDx;
						}
						// next line	

						pSource += sourcePad;
						pD = (RGBTRIPLE*)((char*)pD + destStoreWidth);
					}
				}
			}
		}
		else
		{
			return 7 ;
		}
	}
	else
	{
		return 9 ;
	}
			
    return 0 ;
}

unsigned int CDib::getCornerPixel ( int location )
{
	int byteWidth  = width * bpp/8 ;
	int storeWidth = (byteWidth + 3 ) & ~3 ;
	
	unsigned int val ;

	if ( bpp == 8 )
	{
		BYTE pix = 0 ;

		switch ( location ) 
		{
		case TOP_LEFT_PIXEL :
			pix = *(pPixels + storeWidth*(height - 1 )) ; 
			break ;
		case TOP_RIGHT_PIXEL :
			pix = *(pPixels + width - 1 + storeWidth*(height - 1 )) ;
			break ;
		case BOTTOM_LEFT_PIXEL :
			pix = *(pPixels) ;
			break ;
		case BOTTOM_RIGHT_PIXEL :
			pix = *(pPixels + width - 1 ) ;
			break ;
		}
		val = (unsigned int)pix ;
	}
	else if ( bpp == 16 )
	{
		WORD pix ;

		switch (location)
		{
		case TOP_LEFT_PIXEL:
			pix = *(WORD*)(pPixels + storeWidth*(height - 1 )) ;  
			break ;
		case TOP_RIGHT_PIXEL :
			pix = *(WORD*)(pPixels + (width - 1)*2 + storeWidth*(height - 1 )) ;
			break ;
		case BOTTOM_LEFT_PIXEL :
			pix = *(WORD*)(pPixels) ;
			break ;
		case BOTTOM_RIGHT_PIXEL :
			pix = *(WORD*)(pPixels + (width - 1)*2 ) ;
			break ;
		}
		val = (unsigned int)pix ;
	}
	else if ( bpp == 24 )
	{
		RGBTRIPLE pix ;

		switch (location)
		{
		case TOP_LEFT_PIXEL :
			pix = *(RGBTRIPLE*)(pPixels + storeWidth*(height - 1 )) ;
			break ;
		case TOP_RIGHT_PIXEL :
			pix = *(RGBTRIPLE*)(pPixels + (width - 1)*3 + storeWidth*(height - 1 )) ;
			break ;
		case BOTTOM_LEFT_PIXEL :
			pix = *(RGBTRIPLE*)(pPixels) ;
			break ;
		case BOTTOM_RIGHT_PIXEL :
			pix = *(RGBTRIPLE*)(pPixels + (width - 1)*3 ) ;
			break ;
		}
		val = *(unsigned int *)&pix ;
	}
	else
		val = 0 ;

	return val ;
}

int CDib::draw ( CRect &destRect , CRect &sourceRect , 
				  CDC * pDC , CDC * pMemDC = NULL ,
				  int palOp = 0 , CPalette * pPal = NULL ) 
{
	// copies the dib to the screen. If pMemDC is given, it will use that.
	// If not, it will create one.

	bool madePalette , madeMemDC ;
	
	int  status ;

	madePalette = false ;
	madeMemDC   = false ;

	CPalette * pOldPal ;

	if ( bpp == 8 && palOp != PALOP_NONE )
    {
	    if ( pPal == NULL )
		{
			if ( isPalettized )
				return ERROR_PALETTE_IS_REQUIRED ;

			// create a palette from the RGB colour table,
			// then select and realize it

			pPal = new (CPalette) ;

			status = makePalette ( pPal ) ;

			if ( status != 0 )
			{
				return status ;   // could not make palette
			}
			else
				madePalette = TRUE ;
        }
		pOldPal = pDC->SelectPalette(pPal, (palOp == PALOP_BACKGROUND) ) ; 

		(void)pDC->RealizePalette();
	}
	
	// check that the source rectangle is actually contained in the bitmap.
	// If not, adjust the start and end positions.

	int bmpX , bmpY , bmpW , bmpH ;

	bmpX = sourceRect.left ;
	bmpY = sourceRect.top ;  
	bmpW = sourceRect.right - sourceRect.left ;
	bmpH = sourceRect.bottom - sourceRect.top ;

	if (bmpX == 0 && bmpY == 0 && (bmpW == width)
		&& (bmpH == height) )
	{
		// this is a request to draw the complete bitmap
	}
	else // request for a portion of the bitmap
	{
		if (bmpW >= 0 && bmpH >= 0)
		{
			// simplest case, not dealing with mirror images!

			if (bmpX >= width || bmpY >= height)
			{
				// requested rectangle is outside the bitmap, so don't draw!
				bmpW = 0 ;
				bmpH = 0 ;
			}
			else
			{
				if (bmpW == 0)
					bmpW = width - bmpX ;
				else if (bmpX + bmpW > width)
					bmpW = width - bmpX ;
				if (bmpH == 0)
					bmpH = height - bmpY;
				else if (bmpY + bmpH > height)
					bmpH = height - bmpY ;

				// adjust bmpY to be relative to lower-left corner instead
				// of upper left

				if ( ! isDibSection )
					bmpY = height - bmpY - bmpH ;
			}
		}
		else
		{
			// either the width or the height (or both) is negative
			// to make a mirror image. 

			// no clever defaults here. Assume that the caller is no
			// relying on us to calculate anything!

			// However, we should check that they are valid, and
			// clip if necessary

			if ( isDibSection ) 
			{
				if ( bmpW < 0 ) 
					status = ERROR_BAD_WIDTH_PARAM ;
				else
					status = ERROR_BAD_HEIGHT_PARAM ;
				goto finished ;
			}
			if (bmpX >= width || bmpY >= height)
			{
				// origin is outside the bit map, so draw nothing!

				bmpW = 0 ;
				bmpH = 0 ;
			}
			else
			{
				if (bmpX + bmpW < 0)
					bmpW = 0 - bmpX - 1;
				if (bmpY + bmpH < 0)
					bmpH = 0 - bmpY - 1;

				// adjust the Y origin to be relative to top instead
				// of bottom

				bmpY = height - bmpY - bmpH - 2 ;
			}
		}
	}

	CWnd * pWnd ;

	pWnd = pDC->GetWindow() ;

	CRect * pWinRect ;

	pWinRect = new (CRect) ;

	pWnd->GetClientRect ( pWinRect ) ; // allows checking that the
										// top left position is within
										// the window


	// Only draw the bitmap if part of it will show up in the window!

	int destX , destY , destWidth , destHeight ;

	destX = destRect.left ;
	destY = destRect.top ;
	destWidth = destRect.right - destRect.left ;
	destHeight = destRect.top - destRect.bottom ;

	if ( (destX <= pWinRect->right || destY <= pWinRect->bottom ||
		  destX + destHeight <= pWinRect->right || destY + destWidth <= pWinRect->bottom)
		  && bmpH != 0 && bmpW != 0
	   )
			// destHeight or destWidth may be negative, to specify origin is
			// on bottom or right!
		  
	{
		// bitmap start position is within the window, so draw it!

		if ( ! isDibSection )
		{
			int param = 0 ;

			// work out type of colour table, if 256-colour bmp

			if ( bpp == 8 )
			{
				param = (isPalettized) ? DIB_PAL_COLORS : DIB_RGB_COLORS;
			}

			// if true-colour or hi-colour mode, then no colour table is needed, so
			// the colour-table-type parameter is set to ZERO.

			::StretchDIBits (
				           pDC->m_hDC , 
						   destX, destY,            // destination position
						   (destWidth == 0)?bmpW
										   :destWidth,  // destination width
						   (destHeight == 0)?bmpH
											:destHeight, // destination width
						   bmpX, 
						   bmpY ,  
						   bmpW,
						   bmpH,
						   pPixels,			  // the pixels
						   (LPBITMAPINFO)pBmpInfoHdr,		 // BITMAPINFO
						   param,
						   SRCCOPY);
		}
		else
		{
			if ( pMemDC == NULL )
			{
				pMemDC = new (CDC) ;
			    pMemDC->CreateCompatibleDC(pDC); // we need a memory DC for the DIB section
				madeMemDC = true ;
            }

			CBitmap * pOldBM ;

			pOldBM = (CBitmap*)(pMemDC->SelectObject(hBmp)) ; // select the DIB section
                                                                        
			// To display the DIB section, just BitBlt it onto the screen.
			// This is possible, because the DIB section is a HBITMAP object
			// that can be selected into a Device Context.

	        pDC->BitBlt ( destX, destY, bmpW, bmpH,
	                      pMemDC, bmpX, bmpY, SRCCOPY
	                    );

			//pMemDC->SelectObject(pOldBM);

			if ( madeMemDC ) 
			{
				delete pMemDC ;
			}
		}
	}

	delete pWinRect ;

	status = 0 ;

finished :
	if ( madePalette ) 
	{
		// reselect the old palette, so that we can delete the one we created

		(void)pDC->SelectPalette(pOldPal, TRUE) ;  // select into bg
											       // This allows
		                                           // new palette to be deleted,
		                                            
		pPal->DeleteObject() ;

		delete pPal ;
	}

	return status ;
}

int CDib::maskMerge ( int destXpos , int destYpos ,
		              LPBITMAPINFOHEADER psourceBmpInfoHdr ,
					  BYTE * psourceBmpPixels ,
					  int sourceXpos , int sourceYpos ,
					  int width ,
					  int height ,
					  char * pMask ,
					  int maskWidth ,
					  int maskHeight ) 
{
	// Does a masked copy.
	//
	// pMask points at a string that indicates which bits are to be copied.
	// "0" implies don't copy.
	// "1" implies copy.
	// 
	// The mask is tiled across the source bitmap.
	//

	if ( strlen(pMask) < maskWidth*maskHeight )
		return 24 ;  // mask not long enough!

	LPBITMAPINFOHEADER pdestBmpInfoHdr = this->pBmpInfoHdr ;

	BYTE * pdestBmpPixels = this->pPixels ;

	int	sourceBpp = psourceBmpInfoHdr->biBitCount;
	int   destBpp = pdestBmpInfoHdr->biBitCount ;

	LPBITMAPINFO pdestBmpInfo = (BITMAPINFO *)((char*)pdestBmpInfoHdr);
	LPBITMAPINFO psourceBmpInfo  = (BITMAPINFO *)((char*)psourceBmpInfoHdr);

	if ( sourceXpos < 0 )
	{
		destXpos -= sourceXpos ;
		width    += sourceXpos ;
		sourceXpos = 0 ;
	}
	if ( sourceYpos < 0 )
	{
		destYpos -= sourceYpos ;
		height   += sourceYpos ;
		sourceYpos = 0 ;
	}

	if ( destXpos < 0 )
	{
		sourceXpos -= destXpos ;
		width      += destXpos ;
		destXpos   = 0 ;
	}

	if ( destYpos < 0 )
	{
		sourceYpos -= destYpos ;
		height     += destYpos ;
		destYpos    = 0 ;
	}

	if ( width < 0 || height < 0 )
		return 1 ;

	if ( destXpos >= pdestBmpInfoHdr->biWidth || destYpos >= pdestBmpInfoHdr->biHeight )
	{
		return 1 ;
	}

	// same check for source image

	if ( sourceXpos >= psourceBmpInfoHdr->biWidth || 
		 sourceYpos >= psourceBmpInfoHdr->biHeight )
	{
		return 2 ;
	}

	if (width == 0)
		width = psourceBmpInfoHdr->biWidth;

	if (height == 0)
		height = psourceBmpInfoHdr->biHeight;

	// Make sure not copying beyond the bounds of the input image

	if ( sourceXpos + width >= psourceBmpInfoHdr->biWidth )
		width = psourceBmpInfoHdr->biWidth - sourceXpos ;

	if ( sourceYpos + height >= psourceBmpInfoHdr->biHeight )
		height = psourceBmpInfoHdr->biHeight - sourceYpos ;

	if (destXpos + width >= pdestBmpInfoHdr->biWidth)
		width = pdestBmpInfoHdr->biWidth - destXpos ;

	if (destYpos + height >= pdestBmpInfoHdr->biHeight)
		height = pdestBmpInfoHdr->biHeight - destYpos  ;

		// work out actual width of each line in bitmap (each line gets padded to a
	// 32-bit word boundary)

	// 16 bpp takes 2 bytes per colour, 24 bpp takes 3 bytes

	// work out width of each line in bytes

	int destWidth , destHeight , sourceWidth , sourceHeight ;

	int destByteWidth , sourceByteWidth ;

	int destStoreWidth , destStoreHeight , sourceStoreWidth , sourceStoreHeight ;

	int destPad , sourcePad ;

	BYTE * pDest , * pSource ;

	destWidth = pdestBmpInfoHdr->biWidth ;
	destHeight = pdestBmpInfoHdr->biHeight ;
	sourceWidth = psourceBmpInfoHdr->biWidth ;
	sourceHeight = psourceBmpInfoHdr->biHeight ;

	destByteWidth = (pdestBmpInfoHdr->biWidth * destBpp/8 ) ;
	sourceByteWidth = (psourceBmpInfoHdr->biWidth * sourceBpp/8 ) ;

	destStoreWidth = (destByteWidth + 3 ) & ~3 ;
	sourceStoreWidth = (sourceByteWidth + 3 ) & ~3 ;

	destStoreHeight = pdestBmpInfoHdr->biHeight ;
	sourceStoreHeight = psourceBmpInfoHdr->biHeight;

	// work out number of extra bytes at end of each stored line

	destPad = destStoreWidth - width*destBpp/8 ;
	sourcePad = sourceStoreWidth - width*sourceBpp/8 ;

	pDest = pdestBmpPixels + destStoreWidth*(destStoreHeight - destYpos - height )
		             + destXpos*destBpp/8;

	pSource = psourceBmpPixels + sourceStoreWidth*(sourceStoreHeight - sourceYpos - height ) 
					 + sourceXpos*sourceBpp/8 ;

	if ( destBpp == 8 )
	{
		if ( sourceBpp == 8 )
		{
			// merging 256-colour image into 256-colour image

			BYTE pixel ;

			int row , col ;

			char * pMaskRow = pMask ;

			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pSource++;
					if ( pMaskRow[mCol++] != '0' )
						*(pDest++) = pixel;
					else
						pDest++;

					if ( mCol >= maskWidth )
						mCol = 0 ;
				}
				// next line
				pSource += sourcePad; // = sourceStoreWidth - width
				pDest   += destPad;   // = destStoreWidth - width
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}
		}
		else
		{
			return 3 ;    // can't copy 16 bpp or 24 bpp into 8 bpp
		}
	}
	else if ( destBpp == 16 )
	{
		// copying into a 16-bit image

		if ( sourceBpp == 16 )
		{
			// copying with same colour-depth

			// transparent copying of 16-bit image to a 16-bit image

			WORD pixel ;

			int row , col ;

			WORD * pS = (WORD*)pSource ;
			WORD * pD = (WORD*)pDest ;

			char * pMaskRow = pMask ;
			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pS++;
					if (pMaskRow[mCol++] != '0' )
						*pD++ = pixel;
					else
						pD++;
					if ( mCol >= maskWidth )
						mCol = 0 ;
				}
				// next line

				pS += sourcePad/2;  // we can get away with this because we know
				                    // that 2 pixels exactly fit into each word,
				                    // and there are two bytes per pixel. The 
				                    // byte length of each line is always a multiple
						            // of 4, so everything is evenly divisible.
				pD += destPad/2;
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}
		}
		else if (sourceBpp == 24 ) 
		{
			// copying 24-bit image into 16-bit
			// Need to reduce each pixel from 3 bytes down to 2

			// transparent copying of 24-bit image to a 16-bit image

			RGBTRIPLE pixel  ;

			RGBTRIPLE * pS = (RGBTRIPLE*)pSource ;
			WORD      * pD = (WORD *)pDest ; 

			int row , col ;

			char * pMaskRow = pMask ;

			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pS++;
						
					if ( pMaskRow[mCol++] != '0' )
						*pD++ = ((pixel.rgbtRed>>3)<<10) +
						        ((pixel.rgbtGreen>>3)<<5) +
								 (pixel.rgbtBlue>>3) ;
					else
						pD++;
					if ( mCol >= maskWidth )
						mCol = 0 ;
				}
				// next line

				pD += destPad/2;
			    pS = (RGBTRIPLE*)((char*)pS + sourcePad) ; 
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}
		}
		else if ( sourceBpp == 8 )
		{
			// copying 256-colour image into 16-bit image

			// not allowed if source image is palettized ... hopefully, that will
			// already have been checked

			// transparent copying of 8-bit image to a 16-bit image

			BYTE pixel ;

			WORD      * pD = (WORD *)pDest ; 

			RGBQUAD* pColourTable = (LPRGBQUAD)((char*)psourceBmpInfoHdr 
				                          + sizeof(BITMAPINFOHEADER));

			int row , col ;

			char * pMaskRow = pMask ;
			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pSource++;
					if (pMaskRow[mCol++] != '0')
							*pD++ = (((pColourTable[pixel].rgbRed)>>3)<<10) +
							    (((pColourTable[pixel].rgbGreen)>>3)<<5) +
							    ((pColourTable[pixel].rgbBlue)>>3) ;
					else
						pD++;
					if ( mCol >= maskWidth )
						mCol = 0 ;

				}
				// next line

				pSource += sourcePad ;
				pD += destPad/2 ;
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}
		}
		else
		{
			return 6 ;
		}
	}
	else if ( destBpp == 24 )
	{
		// copying into a 24-bit image

		if ( sourceBpp == 24 )
		{
			// copying with same colour-depth

			// transparent copying of 24-bit image to a 24-bit image

			RGBTRIPLE pixel ;

			int row , col ;

			RGBTRIPLE * pS = (RGBTRIPLE*)pSource ;
			RGBTRIPLE * pD = (RGBTRIPLE*)pDest ;

			char * pMaskRow = pMask ;
			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pS++;
					if ( pMaskRow[mCol++] != '0' )
						*pD++ = pixel;
					else
						pD++;
					if ( mCol >= maskWidth ) 
						mCol = 0 ;
				}
				// next line
	
				pS = (RGBTRIPLE*)((char*)pS + sourcePad);
				pD = (RGBTRIPLE*)((char*)pD + destPad);
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}
		}
		else if (sourceBpp == 16 ) 
		{
			// copying 16-bit image into 24-bit
			// Need to increase each pixel from 2 bytes to 3

			// transparent copying of 16-bit image to a 24-bit image

			WORD pixel ;

			RGBTRIPLE * pD = (RGBTRIPLE*)pDest ;
			WORD      * pS = (WORD *)pSource ; 

			int row , col ;

			char * pMaskRow = pMask ;
			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pS++;
					if ( pMaskRow[mCol++] != '0' )  
					{
						pD->rgbtRed = (pixel & 0x7C00) >> (10-3) ;
						pD->rgbtGreen = (pixel & 0x03E0) >> (5-3) ;
						pD->rgbtBlue = (pixel & 0x001F) << 3 ;
						pD++ ;
						if ( mCol >= maskWidth ) 
							mCol = 0 ;
					}
					else
						pD++ ;
				}
				// next line
	
				pS += sourcePad/2 ;
				pD = (RGBTRIPLE*)((char*)pD + destPad);
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}
		}
		else if ( sourceBpp == 8 )
		{
			// copying 256-colour image into 24-bit image

			// not allowed if source image is palettized ... hopefully, that will
			// already have been checked

			BYTE pixel ;

			RGBTRIPLE      * pD = (RGBTRIPLE *)pDest ; 

			RGBQUAD* pColourTable = (LPRGBQUAD)((char*)psourceBmpInfoHdr 
				                          + sizeof(BITMAPINFOHEADER));

			int row , col ;

			char * pMaskRow = pMask ;
			int mRow = 0 ;
			int mCol = 0 ;

			for (row = 0 ; row < height ; row++)
			{
				for (col = 0 ; col < width ; col++)
				{
					pixel = *pSource++;
					if ( pMaskRow[mCol++] != '0' )  
					{
						pD->rgbtBlue = pColourTable[pixel].rgbBlue ;
						pD->rgbtGreen = pColourTable[pixel].rgbGreen ;
						pD->rgbtRed = pColourTable[pixel].rgbRed ;
						pD++ ;
						if ( mCol >= maskWidth ) 
							mCol = 0 ;
					}
					else
						pD++ ;
				}
				// next line	

				pSource += sourcePad ;
				pD = (RGBTRIPLE*)((char*)pD + destPad);
				if ( ++mRow >= maskHeight )
				{
					mRow = 0 ;
					pMaskRow = pMask ;
				}
				mCol = 0 ;
			}

		}
		else
		{
			return 7 ;
		}
	}
	else
	{
		return 9 ;
	}
			
    return 0 ;
}

int CDib::synchronize ( CDib * pSourceDib )
{
	// check consistency of parameters. 
	//
	// Target bitmap cannot be palettized, as there is no palette to use as a
	// reference.

	if ( this->bpp != 8 || pSourceDib->bpp != 8 )
	{
		return ERROR_NOT_256_COLOURS ;
	}

	if ( this->isPalettized )
	{
		return ERROR_CANNOT_USE_PALETTIZED_DIB ;
	}

	// The base bitmap cannot be palettized either, again because there is no
	// palette to use as a reference.

	if ( pSourceDib->isPalettized )
	{
		return ERROR_CANNOT_USE_PALETTIZED_DIB ;
	}

	LPBITMAPINFOHEADER pbaseBmpInfoHdr , pnewBmpInfoHdr ;

	pbaseBmpInfoHdr = pSourceDib->pBmpInfoHdr ;

	pnewBmpInfoHdr = this->pBmpInfoHdr ;
	

	BITMAPINFO * pnewBmpInfo , * pbaseBmpInfo ;

	BYTE * pbaseBmpPixels , * pnewBmpPixels ;

	pnewBmpInfo  = (BITMAPINFO *)((char*)pnewBmpInfoHdr);

	pnewBmpPixels  = this->pPixels ;

	// check that target contain 256-colour tables

	pbaseBmpInfo = (BITMAPINFO *)((char*)pbaseBmpInfoHdr);

	pbaseBmpPixels = pSourceDib->pPixels ;
	
	// create a palette from the base picture
	
	CPalette  myPal ;

	if ( (pSourceDib->makePalette ( &myPal)) != 0 )
	{
		return ERROR_CANNOT_CREATE_PALETTE ; // could not create base palette
	}
	
	// Get address of new bitmap's colour table

	RGBQUAD* pColourTable ;
		
	pColourTable = (LPRGBQUAD)((char*)pnewBmpInfoHdr + sizeof(BITMAPINFOHEADER));

    // Colour table immediately follows Bit Map Info header

	unsigned int map[256] ; // translation table

	for ( int i = 0 ; i < 256 ; i++ )
	{
		map[i] = myPal.GetNearestPaletteIndex (  
					RGB(pColourTable->rgbRed,
					    pColourTable->rgbGreen,
						pColourTable->rgbBlue
					   )
					);

		pColourTable++ ;
	}

	unsigned int j ;

	int is_1_to_1 ;
		
	is_1_to_1 = 0 ;

	for ( j = 0 ; j < 256 && map[j] == j ; j++ );

	if ( j == 256)
	{
		// The new picture's colour table already directly corresponds
		// to the base picture. There is no need to rearrange pixels.
		// Colour table only needs changing if it is to be converted
		// to PAL colours.

		is_1_to_1 = 1 ;   // it is a 1-to-1 mapping!
	}
	else
	{

		// need to change all the pixel values if not 1-to-1
	
		// change all the new picture's pixels!

		// Work out number of bytes. Each line in the picture is padded
		// out to a 32-bit word boundary.

		int lineWidth ;
		lineWidth = (pnewBmpInfoHdr->biWidth + 3) & ~3;

		int numBytes ;
		numBytes = pnewBmpInfoHdr->biHeight * lineWidth;

		for (j = numBytes ; j > 0 ; j--)
		{
			*pnewBmpPixels = map[*pnewBmpPixels] ;
			pnewBmpPixels++;
		}

		// change the picture's colour table to match the base picture's
		// palette (which will exactly match the base picture's RGB
		// colour table)

		pColourTable = (LPRGBQUAD)((char*)pnewBmpInfoHdr + sizeof(BITMAPINFOHEADER));

		PALETTEENTRY pal[256];
    
		int numRetrieved;

		numRetrieved = myPal.GetPaletteEntries(0, 256, pal);
	
		if (numRetrieved != 256)
		{
			myPal.DeleteObject() ;
			return 2000 + numRetrieved; 
		}
	
		for (j = 0 ; j < 256 ; j++ )
		{
			pColourTable->rgbRed = pal[j].peRed;
			pColourTable->rgbGreen = pal[j].peGreen;
			pColourTable->rgbBlue = pal[j].peBlue;

			pColourTable++;
		}
	}

	myPal.DeleteObject () ; 
	
	return 0 ;

}


int CDib::identityPalette ( CPalette * pPal , CDC * pDC ) 
{
	// converts the palette to an identity palette. It will be
	// selected and realized too.

	int status = 0 ;

    int staticColours ;
    int sysPaletteSize ;

    staticColours = pDC->GetDeviceCaps (NUMCOLORS) ;
    sysPaletteSize = pDC->GetDeviceCaps (SIZEPALETTE);

	if (staticColours <= 0)
	{
		// screen is set to more than 256 colours, and is therefore not a
		// palette device! It is useless to create an identity palette, as
		// there is no hardware palette! The logical palette is already an
		// identity palette, because it has been created based on the
		// bitmap. All we need to do is select and realize the palette, so
		// that Windows can emulate the palette behaviour.

		(void)pDC->SelectPalette(pPal, FALSE);
		(void)pDC->RealizePalette();
	}     
	else
	{
		// need to change the system palette

		CPalette * pOldPal ;

	    if (sysPaletteSize != 256)
		{
			status = ERROR_HW_PALETTE_NOT_256_COLOURS ;
			goto finished ; 
		}
		
		// load our palette into the system palette to get all our colours in.
		// System palette is cleared out first, so that our colours (hopefully) 
		// will get inserted in the same order as we have them.

		// ***Do this even for hi-colour and true-colour modes. The end result 
		// ***will be a palette that includes window's static colours.

		::SetSystemPaletteUse (pDC->m_hDC , SYSPAL_NOSTATIC ) ; // clear them all
		::SetSystemPaletteUse (pDC->m_hDC , SYSPAL_STATIC) ;     // get static ones back

		pOldPal = pDC->SelectPalette(pPal , FALSE) ; // make our palette the foreground palette

		pDC->RealizePalette() ;   // system palette now contains our colours

		(void)pDC->SelectPalette(pOldPal, FALSE) ; // reselect old palette. This frees up our
				                                   // palette, so we can fiddle with it.
    
		// Now get a copy of the system palette, so we can match it back with our palette

		PALETTEENTRY sysPal[256];

		int i ;

		i = ::GetSystemPaletteEntries( pDC->m_hDC, 0, sysPaletteSize, sysPal);

		if ( i != 256 )
		{
			status = ERROR_NOT_256_HW_PALETTE_ENTRIES ;
			goto finished ; // did not get all the sys palette entries!
		}

		// Change our palette so that it exactly matches the system palette

		for (i = 0 ; i < staticColours/2 ; i++ )
			sysPal[i].peFlags = 0 ;

		for ( ; i < sysPaletteSize - staticColours/2 ; i++ )
			sysPal[i].peFlags = PC_NOCOLLAPSE;

		for ( ; i < sysPaletteSize ; i++ )
			sysPal[i].peFlags = 0 ;

		// peFlags are set to PC_NOCOLLAPSE so that when our logical palette is realized, duplicate
		// colours will get mapped to individual palette entries. However, the static colours are
		// always in the system palette, so we *do* need to map to them. This is why they do not
		// get the PC_NOCOLLAPSE flag set.

		pPal->ResizePalette(sysPaletteSize); // being super-safe here, making sure that our palette
				                            // contains 256 entries.
		pPal->SetPaletteEntries (0, sysPaletteSize, sysPal); // map the system palette onto our
					                                         // ...palette now

		// our palette is now an identity palette, exactly matching the system palette!

		// select and realize our palette now. This should not change any system palette
		// entries, because the palette is based on what was already in there! And also
		// because we have not realized any other palette since last realizing this one!

		(void)pDC->SelectPalette ( pPal, FALSE) ;  
		(void)pDC->RealizePalette ();
	}

finished :

	return status ;
}

